==> oauth-login-service-1.11.4-sources.jar: org/clazzes/login/oauth/impl/JWTokenValidator.java <==
/***********************************************************
 * $Id$
 *
 * OAuth Login Services of the clazzes.org project
 * http://www.clazzes.org
 *
 * Created: 02.07.2017
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 ***********************************************************/

package org.clazzes.login.oauth.impl;

import java.io.IOException;
import java.security.InvalidKeyException;
import java.security.MessageDigest;
import java.security.NoSuchAlgorithmException;
import java.security.Signature;
import java.security.SignatureException;
import java.time.Clock;
import java.time.Instant;
import java.util.Arrays;
import java.util.List;
import java.util.Map;

import org.clazzes.login.jbo.common.Algorithm;
import org.clazzes.login.jbo.common.Helpers;
import org.clazzes.login.jbo.jwt.JWKPubKey;
import org.clazzes.login.jbo.jwt.JWToken;
import org.clazzes.login.jbo.jwt.JWTokenClaims;
import org.clazzes.login.jbo.jwt.JWTokenParser;
import org.clazzes.login.oauth.ConfigOptions;
import org.clazzes.login.oauth.DomainConfig;
import org.clazzes.login.oauth.DomainManager;
import org.clazzes.login.oauth.OAuthTokenErrorResponse;
import org.clazzes.login.oauth.OAuthTokenResponse;
import org.clazzes.login.oauth.TokenValidator;
import org.clazzes.util.aop.i18n.Messages;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * An Open ID token validator.
 */
public class JWTokenValidator implements TokenValidator {
    private Clock clock = Clock.systemDefaultZone();
    public void setClock(Clock clock) {
        this.clock = clock;
    }

    private static final Logger log = LoggerFactory.getLogger(JWTokenValidator.class);

    @Override
    public JWToken validateToken(DomainManager domainManager, OAuthTokenResponse response, Messages i18n)
            throws OAuthTokenErrorResponse {

        return this.validateTokenInternal(
            domainManager,
            response,
            i18n,
            domainManager.getDomainConfig().getClientCredentials().getUserName(),
            domainManager.getDomainConfig()
        );
    }

    @Override
    public JWToken validateRobotToken(DomainManager domainManager, OAuthTokenResponse response, Messages i18n)
            throws OAuthTokenErrorResponse {

        return this.validateTokenInternal(
            domainManager,
            response,
            i18n,
            domainManager.getDomainConfig().getRobotCredentials().getUserName(),
            domainManager.getDomainConfig()
        );
}

    protected JWToken validateTokenInternal(DomainManager domainManager,
                                         OAuthTokenResponse response,
                                         Messages i18n,
                                         String clientId,
                                         DomainConfig domainConfig)
            throws OAuthTokenErrorResponse {

        try {

            boolean jwt = "jwt".equals(response.getTokenType()) || domainConfig.getOptions().contains(ConfigOptions.jwtAccessToken);

            String idToken = jwt ? response.getAccessToken() : response.getIdToken();

            if (idToken == null) {
                return null;
            }

            if (log.isDebugEnabled()) {
                log.debug("Validating {} token [{}].",jwt?"JWT":"ID",idToken);
            }

            JWToken token = JWTokenParser.parseJWToken(idToken);
            JWTokenClaims claims = token.getClaimSet();

            if (log.isDebugEnabled()) {
                log.debug("Checking signature on ID Token with issuer [{}] and key ID [{}]...",claims.getIssuer(),token.getKeyId());
            }

            Map<String, Object> openIdConfiguration = domainManager.getOpenIdConfiguration();

            if (openIdConfiguration == null) {

                log.error("OpenID configuration of domain ["+domainManager.getDomain()+"] not loaded while validation Open ID token.");
                throw new OAuthTokenErrorResponse("openid-configuration-not-loaded",i18n);
            }

            String domainIssuer = (String) openIdConfiguration.get("issuer");

            if (domainIssuer == null) {

                log.error("OpenID configuration of domain ["+domainManager.getDomain()+"] does not contain an [issuer] entry while validation Open ID token.");
                throw new OAuthTokenErrorResponse("openid-configuration-invalid",i18n);
            }

            // check issuer
            if (domainIssuer.contains("{tenantid}")) {

                Object tid = claims.getAdditionalClaim("tid");

                if (tid != null) {

                    if (log.isDebugEnabled()) {

                        log.debug("Replacing {tenantid} in domain issuer [{}] with tid [{}]",domainIssuer,tid);
                    }

                    domainIssuer = domainIssuer.replace("{tenantid}",tid.toString());
                }
            }

            if (log.isDebugEnabled()) {

                log.debug("Checking domain issuer [{}] against token issuer [{}].",domainIssuer,claims.getIssuer());
            }

            if (!domainIssuer.equals(claims.getIssuer())) {

                if (jwt && openIdConfiguration.get("access_token_issuer") != null) {

                    String tokenIssuer = (String)openIdConfiguration.get("access_token_issuer");

                    // check token issuer
                    if (tokenIssuer.contains("{tenantid}")) {

                        Object tid = claims.getAdditionalClaim("tid");

                        if (tid != null) {

                            if (log.isDebugEnabled()) {

                                log.debug("Replacing {tenantid} in token issuer [{}] with tid [{}]",tokenIssuer,tid);
                            }

                            tokenIssuer = tokenIssuer.replace("{tenantid}",tid.toString());
                        }
                    }

                    if (log.isDebugEnabled()) {
                        log.debug("Checking domain access_token_issuer [{}] against token issuer [{}].",
                                  tokenIssuer,claims.getIssuer());
                    }

                    if (!tokenIssuer.equals(claims.getIssuer())) {
                        log.error("Domain access_token_issuer [{}] differs from token issuer [{}] while validation Open ID token.",domainIssuer,claims.getIssuer());
                        throw new OAuthTokenErrorResponse("openid-token-invalid-issuer",i18n);
                    }
                }
                else {
                    log.error("Domain issuer [{}] differs from token issuer [{}] while validation Open ID token.",domainIssuer,claims.getIssuer());
                    throw new OAuthTokenErrorResponse("openid-token-invalid-issuer",i18n);
                }
            }

            if (!jwt) {
                String[] audiences = Helpers.parseScope(claims.getAudience());

                if (Arrays.binarySearch(audiences,clientId) < 0) {

                    log.error("Domain clientId [{}] is not contained in the token audience [{}] while validating Open ID token.",
                              clientId,claims.getAudience());
                    throw new OAuthTokenErrorResponse("openid-token-invalid-audience",i18n);
                }
            }

            JWKPubKey pubKey = domainManager.getOpenIdKeys().get(token.getKeyId());

            if (pubKey == null) {
                throw new OAuthTokenErrorResponse("openid-token-unknown-key",i18n);
            }

            Object algorithmsSupported_o = openIdConfiguration.get("id_token_signing_alg_values_supported");

            if (algorithmsSupported_o == null || !(algorithmsSupported_o instanceof List)) {

                log.error("OpenID configuration of domain ["+domainManager.getDomain()+"] does not contain an [id_token_signing_alg_values_supported] entry of type array while validation Open ID token.");
                throw new OAuthTokenErrorResponse("openid-configuration-invalid",i18n);
            }

            @SuppressWarnings("unchecked")
            List<String> algorithmsSupported = (List<String>)algorithmsSupported_o;

            Algorithm selectedAlgorithm = null;

            for (String rawAlgorithm : algorithmsSupported) {

                if (token.getAlgorithm().toString().equals(rawAlgorithm)) {
                    selectedAlgorithm = token.getAlgorithm();
                    break;
                }
            }

            if (selectedAlgorithm == null) {

                log.error("OpenID token algorithm [{}] not contained in [id_token_signing_alg_values_supported] entry of the OpenID configuration of domain [{}].",
                        token.getAlgorithm(),domainManager.getDomain());
                throw new OAuthTokenErrorResponse("openid-token-validation-failed",i18n);
            }

            if (log.isDebugEnabled()) {

                log.debug("Checking token signature using algorithm [{}].",selectedAlgorithm);
            }

            Signature signature = Signature.getInstance(selectedAlgorithm.getJceName());

            signature.initVerify(pubKey.getPubKeyInfo().getPublicKey());

            signature.update(token.getSignaturePayload());

            boolean ok = signature.verify(token.getSignature());

            if (ok) {

                if (log.isDebugEnabled()) {
                    log.debug("Successfully checked signature on ID Token with key ID [{}], claims are [{}].",token.getKeyId(),claims);
                }

                // do not check access token, if we validate a JWT access token.
                String accessToken = jwt ? null : response.getAccessToken();

                if (accessToken != null) {

                    Object atHash = token.getClaimSet().getAdditionalClaim("at_hash");

                    if (atHash == null) {

                        if (!domainManager.getDomainConfig().getOptions().contains(ConfigOptions.lenientAccessTokenCheck)) {

                            log.error("OpenID access token validation failed, no [at_hash] given in claims.");
                            throw new OAuthTokenErrorResponse("access-token-validation-failed",i18n);
                        }
                        else {
                            if (log.isDebugEnabled()) {
                                log.debug("Ignoring missing [at_hash] claim during open ID access token validation, because option [lenientAccessTokenCheck] is given.");
                            }
                        }
                    }
                    else {

                        byte[] raw = accessToken.getBytes("US-ASCII");

                        String hashName = selectedAlgorithm.getHashName();

                        MessageDigest md = MessageDigest.getInstance(hashName);

                        byte[] digest = md.digest(raw);

                        byte[] hashArray = Arrays.copyOf(digest,digest.length/2);

                        byte[] atDigestArray = Helpers.parseBase64(atHash.toString());

                        if (!Arrays.equals(hashArray,atDigestArray)) {

                            log.error("OpenID access token validation failed, because [at_hash] value differs from actual acces token hash.");
                            throw new OAuthTokenErrorResponse("access-token-validation-failed",i18n);
                        }
                        else {
                            if (log.isDebugEnabled()) {
                                log.debug("Correctly validated access token against [at_hash] open ID access token.");
                            }
                        }
                    }
                }

                Long nbf = token.getClaimSet().getNotBefore();
                Long exp = token.getClaimSet().getExpiration();

                long now = Instant.now(this.clock).toEpochMilli();

                if (nbf != null && now < nbf.longValue()) {

                    log.error("OpenID access token is not yet valid, nbf claim was [{}].",Helpers.formatTimeStamp(nbf));
                    throw new OAuthTokenErrorResponse("openid-token-not-yet-valid",i18n);
                }

                if (exp != null && now > exp.longValue()) {

                    log.error("OpenID access token has expired, exp claim was [{}].",Helpers.formatTimeStamp(exp));
                    throw new OAuthTokenErrorResponse("openid-token-expired",i18n);
                }

                log.info("Correctly validated openID token valid from [{}] to [{}]",
                        Helpers.formatTimeStamp(nbf),
                        Helpers.formatTimeStamp(exp) );

                return token;
            }
            else {

                log.error("OpenID token signature validation failed.");
                throw new OAuthTokenErrorResponse("openid-token-validation-failed",i18n);
            }

        } catch (RuntimeException | SignatureException | IOException | NoSuchAlgorithmException | InvalidKeyException e) {

            log.error("Error validating OpenID token",e);

            throw new OAuthTokenErrorResponse("openid-token-validation-failed",i18n);
        }


    }
}

==> oauth-login-service-2.0.3-sources.jar: org/clazzes/login/oauth/impl/JWTokenValidator.java <==
/***********************************************************
 * $Id$
 *
 * OAuth Login Services of the clazzes.org project
 * http://www.clazzes.org
 *
 * Created: 02.07.2017
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 ***********************************************************/

package org.clazzes.login.oauth.impl;

import java.io.IOException;
import java.security.InvalidKeyException;
import java.security.MessageDigest;
import java.security.NoSuchAlgorithmException;
import java.security.Signature;
import java.security.SignatureException;
import java.time.Clock;
import java.time.Instant;
import java.time.OffsetDateTime;
import java.time.ZoneId;
import java.util.Arrays;
import java.util.List;
import java.util.Map;

import org.clazzes.login.jbo.common.Algorithm;
import org.clazzes.login.jbo.common.Helpers;
import org.clazzes.login.jbo.jwt.JWKPubKey;
import org.clazzes.login.jbo.jwt.JWToken;
import org.clazzes.login.jbo.jwt.JWTokenClaims;
import org.clazzes.login.jbo.jwt.JWTokenParser;
import org.clazzes.login.oauth.ConfigOptions;
import org.clazzes.login.oauth.DomainConfig;
import org.clazzes.login.oauth.DomainManager;
import org.clazzes.login.oauth.OAuthTokenErrorResponse;
import org.clazzes.login.oauth.OAuthTokenResponse;
import org.clazzes.login.oauth.TokenValidator;
import org.clazzes.util.aop.i18n.Messages;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * An Open ID token validator.
 */
public class JWTokenValidator implements TokenValidator {
    private Clock clock = Clock.systemDefaultZone();
    public void setClock(Clock clock) {
        this.clock = clock;
    }

    private static final Logger log = LoggerFactory.getLogger(JWTokenValidator.class);

    @Override
    public JWToken validateToken(DomainManager domainManager, OAuthTokenResponse response, Messages i18n)
            throws OAuthTokenErrorResponse {

        return this.validateTokenInternal(
            domainManager,
            response,
            i18n,
            domainManager.getDomainConfig().getClientCredentials().getUserName(),
            domainManager.getDomainConfig()
        );
    }

    @Override
    public JWToken validateRobotToken(DomainManager domainManager, OAuthTokenResponse response, Messages i18n)
            throws OAuthTokenErrorResponse {

        return this.validateTokenInternal(
            domainManager,
            response,
            i18n,
            domainManager.getDomainConfig().getRobotCredentials().getUserName(),
            domainManager.getDomainConfig()
        );
    }

    protected static final String formatTimeStamp(Long millis) {

        if (millis == null) {
            return null;
        }
        else {
            return OffsetDateTime.ofInstant(Instant.ofEpochMilli(millis),ZoneId.systemDefault()).toString();
        }
    }

    protected JWToken validateTokenInternal(DomainManager domainManager,
                                         OAuthTokenResponse response,
                                         Messages i18n,
                                         String clientId,
                                         DomainConfig domainConfig)
            throws OAuthTokenErrorResponse {

        try {

            boolean jwt = "jwt".equals(response.getTokenType()) || domainConfig.getOptions().contains(ConfigOptions.jwtAccessToken);

            String idToken = jwt ? response.getAccessToken() : response.getIdToken();

            if (idToken == null) {
                return null;
            }

            if (log.isDebugEnabled()) {
                log.debug("Validating {} token [{}].",jwt?"JWT":"ID",idToken);
            }

            JWToken token = JWTokenParser.parseJWToken(idToken);
            JWTokenClaims claims = token.getClaimSet();

            if (log.isDebugEnabled()) {
                log.debug("Checking signature on ID Token with issuer [{}] and key ID [{}]...",claims.getIssuer(),token.getKeyId());
            }

            Map<String, Object> openIdConfiguration = domainManager.getOpenIdConfiguration();

            if (openIdConfiguration == null) {

                log.error("OpenID configuration of domain ["+domainManager.getDomain()+"] not loaded while validation Open ID token.");
                throw new OAuthTokenErrorResponse("openid-configuration-not-loaded",i18n);
            }

            String domainIssuer = (String) openIdConfiguration.get("issuer");

            if (domainIssuer == null) {

                log.error("OpenID configuration of domain ["+domainManager.getDomain()+"] does not contain an [issuer] entry while validation Open ID token.");
                throw new OAuthTokenErrorResponse("openid-configuration-invalid",i18n);
            }

            // check issuer
            if (domainIssuer.contains("{tenantid}")) {

                Object tid = claims.getAdditionalClaim("tid");

                if (tid != null) {

                    if (log.isDebugEnabled()) {

                        log.debug("Replacing {tenantid} in domain issuer [{}] with tid [{}]",domainIssuer,tid);
                    }

                    domainIssuer = domainIssuer.replace("{tenantid}",tid.toString());
                }
            }

            if (log.isDebugEnabled()) {

                log.debug("Checking domain issuer [{}] against token issuer [{}].",domainIssuer,claims.getIssuer());
            }

            if (!domainIssuer.equals(claims.getIssuer())) {

                if (jwt && openIdConfiguration.get("access_token_issuer") != null) {

                    String tokenIssuer = (String)openIdConfiguration.get("access_token_issuer");

                    // check token issuer
                    if (tokenIssuer.contains("{tenantid}")) {

                        Object tid = claims.getAdditionalClaim("tid");

                        if (tid != null) {

                            if (log.isDebugEnabled()) {

                                log.debug("Replacing {tenantid} in token issuer [{}] with tid [{}]",tokenIssuer,tid);
                            }

                            tokenIssuer = tokenIssuer.replace("{tenantid}",tid.toString());
                        }
                    }

                    if (log.isDebugEnabled()) {
                        log.debug("Checking domain access_token_issuer [{}] against token issuer [{}].",
                                  tokenIssuer,claims.getIssuer());
                    }

                    if (!tokenIssuer.equals(claims.getIssuer())) {
                        log.error("Domain access_token_issuer [{}] differs from token issuer [{}] while validation Open ID token.",domainIssuer,claims.getIssuer());
                        throw new OAuthTokenErrorResponse("openid-token-invalid-issuer",i18n);
                    }
                }
                else {
                    log.error("Domain issuer [{}] differs from token issuer [{}] while validation Open ID token.",domainIssuer,claims.getIssuer());
                    throw new OAuthTokenErrorResponse("openid-token-invalid-issuer",i18n);
                }
            }

            if (!jwt) {

                List<String> audiences = claims.getAudience();

                if (!audiences.contains(clientId)) {

                    log.error("Domain clientId [{}] is not contained in the token audience [{}] while validating Open ID token.",
                              clientId,audiences);
                    throw new OAuthTokenErrorResponse("openid-token-invalid-audience",i18n);
                }
            }

            JWKPubKey pubKey = domainManager.getOpenIdKeys().get(token.getKeyId());

            if (pubKey == null) {
                throw new OAuthTokenErrorResponse("openid-token-unknown-key",i18n);
            }

            Object algorithmsSupported_o = openIdConfiguration.get("id_token_signing_alg_values_supported");

            if (algorithmsSupported_o == null || !(algorithmsSupported_o instanceof List)) {

                log.error("OpenID configuration of domain ["+domainManager.getDomain()+"] does not contain an [id_token_signing_alg_values_supported] entry of type array while validation Open ID token.");
                throw new OAuthTokenErrorResponse("openid-configuration-invalid",i18n);
            }

            @SuppressWarnings("unchecked")
            List<String> algorithmsSupported = (List<String>)algorithmsSupported_o;

            Algorithm selectedAlgorithm = null;

            for (String rawAlgorithm : algorithmsSupported) {

                if (token.getAlgorithm().toString().equals(rawAlgorithm)) {
                    selectedAlgorithm = token.getAlgorithm();
                    break;
                }
            }

            if (selectedAlgorithm == null) {

                log.error("OpenID token algorithm [{}] not contained in [id_token_signing_alg_values_supported] entry of the OpenID configuration of domain [{}].",
                        token.getAlgorithm(),domainManager.getDomain());
                throw new OAuthTokenErrorResponse("openid-token-validation-failed",i18n);
            }

            if (log.isDebugEnabled()) {

                log.debug("Checking token signature using algorithm [{}].",selectedAlgorithm);
            }

            Signature signature = Signature.getInstance(selectedAlgorithm.getJceName());

            signature.initVerify(pubKey.getPubKeyInfo().getPublicKey());

            signature.update(token.getSignaturePayload());

            boolean ok = signature.verify(token.getSignature());

            if (ok) {

                if (log.isDebugEnabled()) {
                    log.debug("Successfully checked signature on ID Token with key ID [{}], claims are [{}].",token.getKeyId(),claims);
                }

                // do not check access token, if we validate a JWT access token.
                String accessToken = jwt ? null : response.getAccessToken();

                if (accessToken != null) {

                    Object atHash = token.getClaimSet().getAdditionalClaim("at_hash");

                    if (atHash == null) {

                        if (!domainManager.getDomainConfig().getOptions().contains(ConfigOptions.lenientAccessTokenCheck)) {

                            log.error("OpenID access token validation failed, no [at_hash] given in claims.");
                            throw new OAuthTokenErrorResponse("access-token-validation-failed",i18n);
                        }
                        else {
                            if (log.isDebugEnabled()) {
                                log.debug("Ignoring missing [at_hash] claim during open ID access token validation, because option [lenientAccessTokenCheck] is given.");
                            }
                        }
                    }
                    else {

                        byte[] raw = accessToken.getBytes("US-ASCII");

                        String hashName = selectedAlgorithm.getHashName();

                        MessageDigest md = MessageDigest.getInstance(hashName);

                        byte[] digest = md.digest(raw);

                        byte[] hashArray = Arrays.copyOf(digest,digest.length/2);

                        byte[] atDigestArray = Helpers.parseBase64(atHash.toString());

                        if (!Arrays.equals(hashArray,atDigestArray)) {

                            log.error("OpenID access token validation failed, because [at_hash] value differs from actual acces token hash.");
                            throw new OAuthTokenErrorResponse("access-token-validation-failed",i18n);
                        }
                        else {
                            if (log.isDebugEnabled()) {
                                log.debug("Correctly validated access token against [at_hash] open ID access token.");
                            }
                        }
                    }
                }

                Long nbf = token.getClaimSet().getNotBefore();
                Long exp = token.getClaimSet().getExpiration();

                long now = Instant.now(this.clock).toEpochMilli();

                if (nbf != null && now < nbf.longValue()) {

                    log.error("OpenID access token is not yet valid, nbf claim was [{}].",formatTimeStamp(nbf));
                    throw new OAuthTokenErrorResponse("openid-token-not-yet-valid",i18n);
                }

                if (exp != null && now > exp.longValue()) {

                    log.error("OpenID access token has expired, exp claim was [{}].",formatTimeStamp(exp));
                    throw new OAuthTokenErrorResponse("openid-token-expired",i18n);
                }

                log.info("Correctly validated openID token valid from [{}] to [{}]",
                        formatTimeStamp(nbf),
                        formatTimeStamp(exp) );

                return token;
            }
            else {

                log.error("OpenID token signature validation failed.");
                throw new OAuthTokenErrorResponse("openid-token-validation-failed",i18n);
            }

        } catch (RuntimeException | SignatureException | IOException | NoSuchAlgorithmException | InvalidKeyException e) {

            log.error("Error validating OpenID token",e);

            throw new OAuthTokenErrorResponse("openid-token-validation-failed",i18n);
        }


    }
}
